/* vxquery.c
**
** A replacement for `vxdump', `vxdomcheck' and `vxusercheck'
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>

#include "prefs.h"
#include "conf.h"
#include "vxerror.h"
#include "tools.h"
#include "dbif.h"
#include "version.h"
#include "vxopt.h"
#include "exithl.h"
#include "exitcodes.h"
#include "mrmacs.h"
#include "sigwrapper.h"

#include "dump.h"
#include "check.h"
#include "edata.h"

static SIGWRP(sighdl)

/* I prefer complete usage messages, but this message may therefor exceed
** the number of lines in a standard VT100 terminal.
*/
void usage (void)
{
    fprintf (stderr, "\nUsage: %s [<options>] <querymode> [<querymode args>]"
		     "\n       %s -h"
		     "\n       %s -V\n"
		     "\n   Options:"
		     "\n      -f <configfile> (default: %s)"
		     "\n      -o <outputfile>\n"
		     "\n   Query-Modes:"
		     "\n      a[ll]              - dump the complete database"
		     "\n      d[omain] <domain> [r[ecreate]]"
		     "\n                         - dump the domain <domain>"
		     "\n      fo[lder] <domain>  - print a domain's folder "
		     "\n                           directory"
		     "\n      fo[lder] <emaddr>  - print a user's folder name"
		     "\n      s[pecial]          - dump all special entries"
		     "\n      s[pecial] IPTRANS  - dump all IP translations"
		     "\n      s[pecial] REDIRECT - dump all POP3-redirects"
		     "\n      s[pecial] IDENTITY - dump the database identity "
		     "record"
		     "\n      s[pecial] VXTS     - dump the VXTS-record"
		     "\n      e[xists] <domain> [v[erbose]]"
		     "\n                         - return with a zero exit "
		     "code if <domain> exists"
		     "\n                           and with EX_NOHOST otherwise"
		     "\n      e[xists] <emaddr> [v[erbose]]"
		     "\n                         - return with a zero exit "
		     "code if <emaddr> exists"
		     "\n                           and with EX_NOUSER otherwise"
		     "\n      p[rofile] <domain> - print the domain profile "
		     "(filter path,"
		     "\n                           user limits, folder limits)"
		     " for <domain>"
		     "\n      fi[lter] <domain>  - print a domain's e-mail "
		     "filter name"
		     "\n      l[imits] <domain>  - print the default maximum "
		     "message/folder sizes"
		     "\n                           for <domain>"
		     "\n      l[imits] <emaddr>  - print the maximum message/"
		     "folder sizes "
		     "\n                           for <emaddr>"
		     "\n      u[sers] <domain>   - print the current/maximum "
		     "number of users"
		     "\n                           in <domain>\n\n",
		     progname, progname, progname, CONFIGFILE);
    exit (EX_USAGE);
}

/* Forward declarations for my query functions */

static int vxq_all (int argc, char *argv[]);
static int vxq_domain (int argc, char *argv[]);
static int vxq_exists (int argc, char *argv[]);
static int vxq_filter (int argc, char *argv[]);
static int vxq_folder (int argc, char *argv[]);
static int vxq_limits (int argc, char *argv[]);
static int vxq_profile (int argc, char *argv[]);
static int vxq_special (int argc, char *argv[]);
static int vxq_users (int argc, char *argv[]);

/* `vxquery' selects it's actions by using the following structure array */

typedef int (*qfunc_t) (int argc, char *argv[]);
struct _qe {
    char *name;
    qfunc_t do_query;
    short minargs, maxargs;
};

static struct _qe queries[] = {
    { "all",     vxq_all,     0, 0 },
    { "domain",  vxq_domain,  1, 2 },
    { "exists",  vxq_exists,  1, 2 },
    { "filter",  vxq_filter,  1, 1 },
    { "folder",  vxq_folder,  1, 1 },
    { "limits",  vxq_limits,  1, 1 },
    { "profile", vxq_profile, 1, 1 },
    { "special", vxq_special, 0, 1 },
    { "users",   vxq_users,   1, 1 },
    { (char *) NULL, (qfunc_t) 0, 0, 0 }
}, *qu;

static __inline__ int is_pfx (char *q, char *s)
{
    size_t lq = strlen (q), ls = strlen (s);
    if (lq > ls) { return 0; }
    return ! lccmp (q, s, lq);
}

DBIF_FILE userdb = (DBIF_FILE) 0;
/* */
int main (int argc, char *argv[])
{
    char *qm, *qn;
    int ix, res, ma;
    int oc, o_set = 0, ofd;

    vxopt (&argc, argv, usage);
    reset_umask();
    SIGNAL (SIGPIPE, sighdl);

    while ((oc = getopt (argc, argv, "o:")) != -1) {
	switch (oc) {
	    case 'o':
		if (o_set) {
		    fprintf (stderr, "%s: duplicate `-o' option ignored!\n",
				     progname);
		    continue;
		}
		ofd = open (optarg, O_CREAT|O_TRUNC|O_WRONLY, prefs.fperm);
		if (ofd < 0) {
		    vxerror (EX_OSERR, errno, "writing output file failed");
		}
		dup2 (ofd, 1);
		close (ofd);
		write (1, "# Output from `", 15);
		write (1, progname, strlen(progname));
		write (1, "'\n", 2);
		break;
	    default:
		fprintf (stderr,
			 "%s: illegal option `%s'; use `%s -h' for help!\n",
			 progname, argv[optind], progname);
		exit (1);
	}
    }
    /* The next step is a bit dirty, but `so what'... */
    if (optind > 0) { argv = &argv[optind - 1]; argc = argc - optind + 1; }

    /* Get the query mode */
    if (!strcmp (progname, "vxdump")) {
	qm = "all";
	ma = 1;
#if 0
    } else if (!strcmp (progname, "vxdomcheck")) {
	qm = "exists";
	ma = 1;
#endif
    } else {
	if (argc < 2) { usage (); }
	qm = argv[1];
	ma = 2;
    }

    /* Read the configuration */
    getconf ();
    get_ug ();

    dbif_reset (&userdb, prefs.userdb_name);
    ifnull (userdb) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__,
		    "opening userdb failed: %s\n", dbif_errormsg ());
    }

    /* select the action */
    for (ix = 0; queries[ix].name; ++ix) {
	/* shorthands */
	qu = &queries[ix];
	qn = qu->name;

	/* i select the corresponding action if the supplied querymode
	** argument is a prefix of the complete querymode name as given in
	** the structure array above
	*/
	if (is_pfx (qm, qn)) {
	    /* check if the query has the correct number of arguments */
	    if (argc - ma < qu->minargs || argc - ma > qu->maxargs) {
		fprintf (stderr, "%s: wrong number of arguments!\n", progname);
		exit (EX_USAGE);
	    }
	    /* the return value if the query function is the return-(exit-)
	    ** code of the program
	    */
	    res = qu->do_query (argc - ma, &argv[ma]);
	    dbif_close (userdb);
	    return res;
	}
    }

    /* If the program flow leads to this point, you have supplied an illegal
    ** (unknown) query.
    */
    dbif_close (userdb);
    fprintf (stderr, "%s: unknown querymode `%s'!\n", progname, argv[1]);
    exit (EX_USAGE);
}

/* This special signal handler is required because temporary files are
** not deleted in the case of a SIGPIPE signal, so this must be done
** manually before exiting.
*/
static void sighdl (int sig)
{
    fprintf (stderr, "Signal %d found - cleaning up ...\n", sig);
    /* clean up - i.e. delete any temporary files, ... */
    run_exithl (NULL);
#if 1
    /* reset the signal handler to the default */
    signal (sig, SIG_DFL);
#endif
    /* re-send the signal */
    kill (getpid (), sig);
}

/* A short helper function which checks if you are the SUPERUSER and aborts
** the program otherwise (some query modes lead to the output of sensible
** data and so can only be done by the SUPERUSER).
*/
static void check_permission (char *querymode)
{
    if (getuid () == 0) { return; }
    fprintf (stderr, "%s: you have no permission to query `%s'!\n",
		     progname,
		     querymode);
    exit (EX_NOPERM);
}

/* set the correct user/group for the `vxmail' system (and database) */
static void setvxuser (void)
{
    setgid (prefs.vxgid);
    setreuid (getuid(), prefs.vxuid);
}

/* this (query mode) function replaces the `vxdump' program */
static int vxq_all (int ac, char *av[])
{
    check_permission ("all");
    setvxuser ();
    return dump_database (userdb);
}

/* this (query mode) function is an extension to the `vxdump' program.
** It dump's only a specified domain (header and entries).
*/
static int vxq_domain (int ac, char *av[])
{
    int recreate = (ac > 1 && is_pfx (av[1], "recreate"));
    char *domain = av[0];
    check_permission ("domain");	/* only `root' can perform this query */
    setvxuser ();
    return dump_domain (userdb, domain, recreate);
}

/* this (query mode) function replaces the `vxdomcheck' and `vxusercheck'
** programs.
*/
static int vxq_exists (int ac, char *av[])
{
    int verbose = (ac > 1 && is_pfx (av[1], "verbose"));
    char *entry = av[0];
    setvxuser ();
    return check_entry (userdb, entry, verbose);
}

/* this (query mode) function prints the name of the filter for a specified
** domain. It prints nothing and returns EX_NOINPUT if the specified domain
** doesn't have a filter
*/
static int vxq_filter (int ac, char *av[])
{
    char *domain = av[0];
    setvxuser ();
    return print_filter (userdb, domain);
}

/* print the path of the folder directory for a specified domain
** or the path of the folder for a specified user
*/
static int vxq_folder (int ac, char *av[])
{
    char *entry = av[0];
    setvxuser ();
    return print_folder (userdb, entry);
}

/* print the default folder limits for users of a specified domain or
** the (current) folder limits of a specified user
*/
static int vxq_limits (int ac, char *av[])
{
    char *entry = av[0];
    check_permission ("limits");	/* only `root' can perform this query */
    setvxuser ();
    return print_limits (userdb, entry);
}

/* print the filter pathname, the user limits and the default folder limits of
** a specified domain
*/
static int vxq_profile (int ac, char *av[])
{
    char *domain = av[0];
    check_permission ("profile");	/* only `root' can perform this query */
    setvxuser ();
    return print_profile (userdb, domain);
}

/* print (all or the entries specified by type of) the special entries of
** the user database
*/
static int vxq_special (int ac, char *av[])
{
    char *sptype = (ac > 0 ? av[0] : NULL);
    check_permission (sptype ? sptype : "special");
					/* only `root' can perform this query */
    setvxuser ();
    return dump_specials (userdb, sptype);
}

/* print the current/maximum number of users a specified domain */
static int vxq_users (int ac, char *av[])
{
    char *domain = av[0];
    check_permission ("users");		/* only `root' can perform this query */
    setvxuser ();
    return print_nusers (userdb, domain);
}
